"""Run all the examples (except for benchmark.py). Just used as some kind of
functional test... If no warning / errors is output, it should be fine.
"""

# flake8: noqa

import os
import sys
import warnings

# redirect stout to /dev/null to avoid printing
sys.stdout = open(os.devnull, "w")

import baselines_conf
import basic_usage
import evaluate_on_trainset
import generate_grid_search_cv_results_example
import grid_search_usage
import k_nearest_neighbors
import load_custom_dataset
import load_custom_dataset_predefined_folds
import load_from_dataframe
import precision_recall_at_k
import predict_ratings
import serialize_algorithm
import similarity_conf
import split_data_for_unbiased_estimation
import top_n_recommendations
import train_test_split
import use_cross_validation_iterators
from building_custom_algorithms import (
    mean_rating_user_item,
    most_basic_algorithm,
    most_basic_algorithm2,
    with_baselines_or_sim,
)



"""
This module descibes how to load a custom dataset from a single file.

As a custom dataset we will actually use the movielens-100k dataset, but act as
if it were not built-in.
"""

import os

from surprise import BaselineOnly, Dataset, Reader
from surprise.model_selection import cross_validate

# path to dataset file
file_path = os.path.expanduser("~/.surprise_data/ml-100k/ml-100k/u.data")

# As we're loading a custom dataset, we need to define a reader. In the
# movielens-100k dataset, each line has the following format:
# 'user item rating timestamp', separated by '\t' characters.
reader = Reader(line_format="user item rating timestamp", sep="\t")

data = Dataset.load_from_file(file_path, reader=reader)

# We can now use this dataset as we please, e.g. calling cross_validate
cross_validate(BaselineOnly(), data, verbose=True)



import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, setup

"""
Prior to relying on PEP517/518 and using pyproject.toml, this setup.py used to
be an unintelligible mess. The main reason being that there were no clear
distinction between run-time and build-time dependencies, and since we didn't
want to make Cython a run-time dep, we had to enable a way to install the sdist
from the .c files instead of from the .pyx file.
Anyways. Now Cython is a build-time dep, not a run-time dep, since installing
from the sdist happens in an isolated env.

Creating the sdist still involves compiling the .pyx into .c because we're
executing this file. This is unnecessary but it doesn't matter. The .c files are
excluded from the sdist (in MANIFEST.in) anyway.

Release instruction:

Update changelog and contributors list. 

Basic local checks:
- tests run correctly
- doc compiles without warning (make clean first).

Check that the latest RTD build was OK: https://readthedocs.org/projects/surprise/builds/

Change __version__ in __init__.py to new version name. Also update the hardcoded
version in build_sdist.yml, otherwise the GA jobs will fail.

The sdist is built on Python 3.8. It should be installable from all Python
versions.
- check the sdist building process. It will (unnecessarily) compily the .pyx
  files and the .c files should be excluded from the archive.
- check the install jobs. This will compile the .pyx files again as well as the
  .c files. Look for compilation warnings.
- check test jobs for warnings etc.

Download the sdist from the CI job then upload it to test pypi: twine upload
blabla.tar.gz -r testpypi

Check that install works on testpypi, then upload to pypi and check again.
to install from testpypi:
pip install --index-url https://test.pypi.org/simple/ --extra-index-url https://pypi.org/simple scikit-surprise  # noqa
Doesn't hurt to check that the tests pass after installing from testpypi.

If not already done, sync gh-pages with the master's README

push new release tag on github (commit last changes first if needed):
    git tag vX.Y.Z
    git push --tags

Check that RTD has updated 'stable' to the new release (may take a while).

In the mean time, upload to conda:
    - Compute SHA256 hash of the new .tar.gz archive (or check it up on PyPI)
    - update recipe/meta.yaml on feedstock fork consequently (only version and
      sha should be changed.  Maybe add some import tests).
    - Push changes, Then open pull request on conda-forge feedstock and merge it
      when all checks are OK. Access the conda-forge feedstock it by the link on
      GitHub 'forked from blah blah'.
    - Check on https://anaconda.org/conda-forge/scikit-surprise that new
      version is available for all platforms.

Then, maybe, celebrate.
"""

# This prevents Cython from using deprecated numpy C APIs
define_macros = [("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]

# We're using numpy C APIs in our Cython code so Cython will generate C code
# that requires the numpy headers. We need to tell the compiler where to find
# those headers.
# If you remove this and compilation still works, don't get fooled: it's
# probably only because the numpy headers are available in the default locations
# like /usr/include/numpy/ so they get found. But that wouldn't necessarily be
# the case on other users' machines building the sdist.
include_dirs = [np.get_include()]

extensions = [
    Extension(
        # name is where the .so will be placed, i.e. where the module will be
        # importable from.
        name="surprise.similarities",
        sources=["surprise/similarities.pyx"],
        include_dirs=include_dirs,
        define_macros=define_macros,
    ),
    Extension(
        name="surprise.prediction_algorithms.matrix_factorization",
        sources=["surprise/prediction_algorithms/matrix_factorization.pyx"],
        include_dirs=include_dirs,
        define_macros=define_macros,
    ),
    Extension(
        name="surprise.prediction_algorithms.optimize_baselines",
        sources=["surprise/prediction_algorithms/optimize_baselines.pyx"],
        include_dirs=include_dirs,
        define_macros=define_macros,
    ),
    Extension(
        name="surprise.prediction_algorithms.slope_one",
        sources=["surprise/prediction_algorithms/slope_one.pyx"],
        include_dirs=include_dirs,
        define_macros=define_macros,
    ),
    Extension(
        name="surprise.prediction_algorithms.co_clustering",
        sources=["surprise/prediction_algorithms/co_clustering.pyx"],
        include_dirs=include_dirs,
        define_macros=define_macros,
    ),
]

extensions = cythonize(
    extensions,
    compiler_directives={
        "language_level": 3,
        "boundscheck": False,
        "wraparound": False,
        "initializedcheck": False,
        "nonecheck": False,
    },
)

setup(ext_modules=extensions)



"""
Module for testing prediction algorithms.
"""

import os

import pytest
from surprise import (
    accuracy,
    BaselineOnly,
    CoClustering,
    Dataset,
    KNNBaseline,
    KNNBasic,
    KNNWithMeans,
    KNNWithZScore,
    NMF,
    NormalPredictor,
    Reader,
    SlopeOne,
    SVD,
    SVDpp,
)
from surprise.model_selection import train_test_split


def test_unknown_user_or_item(toy_data):
    """Ensure that all algorithms act gracefully when asked to predict a rating
    of an unknown user, an unknown item, and when both are unknown.
    """

    trainset = toy_data.build_full_trainset()

    klasses = (
        NormalPredictor,
        BaselineOnly,
        KNNBasic,
        KNNWithMeans,
        KNNBaseline,
        SVD,
        SVDpp,
        NMF,
        SlopeOne,
        CoClustering,
        KNNWithZScore,
    )
    for klass in klasses:
        algo = klass()
        algo.fit(trainset)
        algo.predict("user0", "unknown_item", None)
        algo.predict("unknown_user", "item0", None)
        algo.predict("unknown_user", "unknown_item", None)

    # unrelated, but test the fit().test() one-liner:
    trainset, testset = train_test_split(toy_data, test_size=2)
    for klass in klasses:
        algo = klass()
        algo.fit(trainset).test(testset)


def test_knns(u1_ml100k, pkf):
    """Ensure the k and min_k parameters are effective for knn algorithms."""

    # Actually, as KNNWithMeans and KNNBaseline have back up solutions for when
    # there are not enough neighbors, we can't really test them...
    klasses = (KNNBasic,)  # KNNWithMeans, KNNBaseline)

    k, min_k = 20, 5
    for klass in klasses:
        algo = klass(k=k, min_k=min_k)
        for trainset, testset in pkf.split(u1_ml100k):
            algo.fit(trainset)
            predictions = algo.test(testset)
            for pred in predictions:
                if not pred.details["was_impossible"]:
                    assert min_k <= pred.details["actual_k"] <= k


def test_nearest_neighbors():
    """Ensure the nearest neighbors are different when using user-user
    similarity vs item-item."""

    reader = Reader(
        line_format="user item rating", sep=" ", skip_lines=3, rating_scale=(1, 5)
    )

    data_file = os.path.dirname(os.path.realpath(__file__)) + "/custom_train"
    data = Dataset.load_from_file(data_file, reader)
    trainset = data.build_full_trainset()

    algo_ub = KNNBasic(sim_options={"user_based": True})
    algo_ub.fit(trainset)
    algo_ib = KNNBasic(sim_options={"user_based": False})
    algo_ib.fit(trainset)
    assert algo_ub.get_neighbors(0, k=10) != algo_ib.get_neighbors(0, k=10)


@pytest.mark.parametrize(
    "algo, expected_rmse",
    [
        (BaselineOnly(), 1.0268524031297395),
        (KNNBasic(sim_options={"name": "cosine"}), 1.1495072745332722),
        (KNNBasic(sim_options={"name": "msd"}), 1.1337265249554591),
        (KNNBasic(sim_options={"name": "pearson"}), 1.1218719178944856),
        (KNNBasic(sim_options={"name": "pearson_baseline"}), 1.1242139598121241),
        (KNNWithMeans(), 1.1043129441881696),
        (KNNBaseline(), 1.0700718041752253),
        (KNNWithZScore(), 1.11179436167853),
        (SVD(random_state=0), 1.0077323320656948),
        (SVDpp(cache_ratings=False, random_state=0), 1.00284553561452),
        (SVDpp(cache_ratings=True, random_state=0), 1.00284553561452),
        (NMF(random_state=0), 1.0865370266372372),
        (SlopeOne(), 1.1559939123891685),
        (CoClustering(random_state=0), 1.0841941385276614),
    ],
)
def test_sanity_checks(u1_ml100k, pkf, algo, expected_rmse):
    """Basic sanity checks for all algorithms: check that RMSE stays the same."""

    trainset, testset = next(pkf.split(u1_ml100k))
    algo.fit(trainset)
    predictions = algo.test(testset)
    assert not all(pred.details["was_impossible"] for pred in predictions)
    assert accuracy.rmse(predictions, verbose=False) == expected_rmse



"""
Module for testing the model_selection.search module.
"""

import os

import numpy as np
import pytest
from scipy.stats import randint, uniform
from surprise import Dataset, Reader, SVD
from surprise.model_selection import (
    cross_validate,
    GridSearchCV,
    KFold,
    PredefinedKFold,
    RandomizedSearchCV,
)


# Tests for GridSearchCV class
def test_basesearchcv_parse_options():
    """Check that _parse_options explodes nested dictionaries."""
    param_grid = {
        "bsl_options": {"method": ["als", "sgd"], "reg": [1, 2]},
        "k": [2, 3],
        "sim_options": {
            "name": ["msd", "cosine"],
            "min_support": [1, 5],
            "user_based": [False],
        },
    }
    gs = GridSearchCV(SVD, param_grid)
    parsed_params = gs.param_grid
    assert len(parsed_params["sim_options"]) == 4  # 2 method x 2 reg
    assert len(parsed_params["bsl_options"]) == 4  # 2 method x 2 reg


def test_gridsearchcv_parameter_combinations():
    """Make sure that parameter_combinations attribute is correct (has correct
    size).  Dict parameters like bsl_options and sim_options require special
    treatment in the param_grid argument. We here test both in one shot with
    KNNBaseline."""

    param_grid = {
        "bsl_options": {"method": ["als", "sgd"], "reg": [1, 2]},
        "k": [2, 3],
        "sim_options": {
            "name": ["msd", "cosine"],
            "min_support": [1, 5],
            "user_based": [False],
        },
    }

    gs = GridSearchCV(SVD, param_grid)
    assert len(gs.param_combinations) == 32


def test_gridsearchcv_best_estimator(u1_ml100k):
    """Ensure that the best estimator is the one giving the best score (by
    re-running it)"""

    param_grid = {
        "n_epochs": [5],
        "lr_all": [0.002, 0.005],
        "reg_all": [0.4, 0.6],
        "n_factors": [1],
        "init_std_dev": [0],
    }
    gs = GridSearchCV(
        SVD, param_grid, measures=["mae"], cv=PredefinedKFold(), joblib_verbose=100
    )
    gs.fit(u1_ml100k)
    best_estimator = gs.best_estimator["mae"]

    # recompute MAE of best_estimator
    mae = cross_validate(
        best_estimator, u1_ml100k, measures=["MAE"], cv=PredefinedKFold()
    )["test_mae"]

    assert mae == gs.best_score["mae"]


def test_gridsearchcv_same_splits():
    """Ensure that all parameter combinations are tested on the same splits (we
    check their RMSE scores are the same once averaged over the splits, which
    should be enough)."""

    data_file = os.path.join(os.path.dirname(__file__), "./u1_ml100k_test")
    data = Dataset.load_from_file(data_file, reader=Reader("ml-100k"))
    kf = KFold(3, shuffle=True, random_state=4)

    # all RMSE should be the same (as param combinations are the same)
    param_grid = {
        "n_epochs": [5],
        "lr_all": [0.2, 0.2],
        "reg_all": [0.4, 0.4],
        "n_factors": [5],
        "random_state": [0],
    }
    gs = GridSearchCV(SVD, param_grid, measures=["RMSE"], cv=kf, n_jobs=1)
    gs.fit(data)

    rmse_scores = [m for m in gs.cv_results["mean_test_rmse"]]
    assert len(set(rmse_scores)) == 1  # assert rmse_scores are all equal

    # Note: actually, even when setting random_state=None in kf, the same folds
    # are used because we use product(param_comb, kf.split(...)). However, it's
    # needed to have the same folds when calling fit again:
    gs.fit(data)
    rmse_scores += [m for m in gs.cv_results["mean_test_rmse"]]
    assert len(set(rmse_scores)) == 1  # assert rmse_scores are all equal


def test_gridsearchcv_cv_results():
    """Test the cv_results attribute"""

    f = os.path.join(os.path.dirname(__file__), "./u1_ml100k_test")
    data = Dataset.load_from_file(f, Reader("ml-100k"))
    kf = KFold(3, shuffle=True, random_state=4)
    param_grid = {
        "n_epochs": [5],
        "lr_all": [0.2, 0.4],
        "reg_all": [0.4, 0.6],
        "n_factors": [5],
        "random_state": [0],
    }
    gs = GridSearchCV(
        SVD,
        param_grid,
        measures=["RMSE", "mae", "fcp"],
        cv=kf,
        return_train_measures=True,
    )
    gs.fit(data)

    # test keys split*_test_rmse, mean and std dev.
    assert gs.cv_results["split0_test_rmse"].shape == (4,)  # 4 param comb.
    assert gs.cv_results["split1_test_rmse"].shape == (4,)  # 4 param comb.
    assert gs.cv_results["split2_test_rmse"].shape == (4,)  # 4 param comb.
    assert gs.cv_results["mean_test_rmse"].shape == (4,)  # 4 param comb.
    assert np.allclose(
        gs.cv_results["mean_test_rmse"],
        np.mean(
            [
                gs.cv_results["split0_test_rmse"],
                gs.cv_results["split1_test_rmse"],
                gs.cv_results["split2_test_rmse"],
            ],
            axis=0,
        ),
    )
    assert np.allclose(
        gs.cv_results["std_test_rmse"],
        np.std(
            [
                gs.cv_results["split0_test_rmse"],
                gs.cv_results["split1_test_rmse"],
                gs.cv_results["split2_test_rmse"],
            ],
            axis=0,
        ),
    )

    # test keys split*_train_mae, mean and std dev.
    assert gs.cv_results["split0_train_rmse"].shape == (4,)  # 4 param comb.
    assert gs.cv_results["split1_train_rmse"].shape == (4,)  # 4 param comb.
    assert gs.cv_results["split2_train_rmse"].shape == (4,)  # 4 param comb.
    assert gs.cv_results["mean_train_rmse"].shape == (4,)  # 4 param comb.
    assert np.allclose(
        gs.cv_results["mean_train_rmse"],
        np.mean(
            [
                gs.cv_results["split0_train_rmse"],
                gs.cv_results["split1_train_rmse"],
                gs.cv_results["split2_train_rmse"],
            ],
            axis=0,
        ),
    )
    assert np.allclose(
        gs.cv_results["std_train_rmse"],
        np.std(
            [
                gs.cv_results["split0_train_rmse"],
                gs.cv_results["split1_train_rmse"],
                gs.cv_results["split2_train_rmse"],
            ],
            axis=0,
        ),
    )

    # test fit and train times dimensions.
    assert gs.cv_results["mean_fit_time"].shape == (4,)  # 4 param comb.
    assert gs.cv_results["std_fit_time"].shape == (4,)  # 4 param comb.
    assert gs.cv_results["mean_test_time"].shape == (4,)  # 4 param comb.
    assert gs.cv_results["std_test_time"].shape == (4,)  # 4 param comb.

    assert gs.cv_results["params"] is gs.param_combinations

    # assert that best parameter in gs.cv_results['rank_test_measure'] is
    # indeed the best_param attribute.
    best_index = np.argmin(gs.cv_results["rank_test_rmse"])
    assert gs.cv_results["params"][best_index] == gs.best_params["rmse"]
    best_index = np.argmin(gs.cv_results["rank_test_mae"])
    assert gs.cv_results["params"][best_index] == gs.best_params["mae"]
    best_index = np.argmin(gs.cv_results["rank_test_fcp"])
    assert gs.cv_results["params"][best_index] == gs.best_params["fcp"]


def test_gridsearchcv_refit(u1_ml100k):
    """Test refit function of GridSearchCV."""

    data_file = os.path.join(os.path.dirname(__file__), "./u1_ml100k_test")
    data = Dataset.load_from_file(data_file, Reader("ml-100k"))

    param_grid = {
        "n_epochs": [5],
        "lr_all": [0.002, 0.005],
        "reg_all": [0.4, 0.6],
        "n_factors": [2],
    }

    # assert gs.fit() and gs.test will use best estimator for mae (first
    # appearing in measures)
    gs = GridSearchCV(SVD, param_grid, measures=["mae", "rmse"], cv=2, refit=True)
    gs.fit(data)
    gs_preds = gs.test(data.construct_testset(data.raw_ratings))
    mae_preds = gs.best_estimator["mae"].test(data.construct_testset(data.raw_ratings))
    assert gs_preds == mae_preds

    # assert gs.fit() and gs.test will use best estimator for rmse
    gs = GridSearchCV(SVD, param_grid, measures=["mae", "rmse"], cv=2, refit="rmse")
    gs.fit(data)
    gs_preds = gs.test(data.construct_testset(data.raw_ratings))
    rmse_preds = gs.best_estimator["rmse"].test(
        data.construct_testset(data.raw_ratings)
    )
    assert gs_preds == rmse_preds
    # test that predict() can be called
    gs.predict(2, 4)

    # assert test() and predict() cannot be used when refit is false
    gs = GridSearchCV(SVD, param_grid, measures=["mae", "rmse"], cv=2, refit=False)
    gs.fit(data)
    with pytest.raises(ValueError):
        gs_preds = gs.test(data.construct_testset(data.raw_ratings))
    with pytest.raises(ValueError):
        gs.predict("1", "2")

    # test that error is raised if used with load_from_folds
    gs = GridSearchCV(SVD, param_grid, measures=["mae", "rmse"], cv=2, refit=True)
    with pytest.raises(ValueError):
        gs.fit(u1_ml100k)


# Tests for RandomizedSearchCV


def test_randomizedsearchcv_parameter_combinations_all_lists():
    """Ensure the parameter_combinations attribute populates correctly by
    checking its length."""
    param_distributions = {
        "bsl_options": {"method": ["als", "sgd"], "reg": [1, 2]},
        "k": [2, 3],
        "sim_options": {
            "name": ["msd", "cosine"],
            "min_support": [1, 5],
            "user_based": [False],
        },
    }
    rs = RandomizedSearchCV(SVD, param_distributions, n_iter=10)
    assert len(rs.param_combinations) == 10


def test_randomizedsearchcv_parameter_combinations_with_distribution():
    """Ensure the parameter_combinations attribute populates correctly by
    checking its length."""
    param_distributions = {
        "bsl_options": {"method": ["als", "sgd"], "reg": [1, 2]},
        "k": randint(2, 4),  # min inclusive, max exclusive
        "sim_options": {
            "name": ["msd", "cosine"],
            "min_support": [1, 5],
            "user_based": [False],
        },
    }
    rs = RandomizedSearchCV(SVD, param_distributions, n_iter=10)
    assert len(rs.param_combinations) == 10


def test_randomizedsearchcv_best_estimator(u1_ml100k):
    """Ensure that the best estimator is the one that gives the best score (by
    re-running it)"""

    param_distributions = {
        "n_epochs": [5],
        "lr_all": uniform(0.002, 0.003),
        "reg_all": uniform(0.04, 0.02),
        "n_factors": [1],
        "init_std_dev": [0],
    }
    rs = RandomizedSearchCV(
        SVD,
        param_distributions,
        measures=["mae"],
        cv=PredefinedKFold(),
        joblib_verbose=100,
    )
    rs.fit(u1_ml100k)
    best_estimator = rs.best_estimator["mae"]

    # recompute MAE of best_estimator
    mae = cross_validate(
        best_estimator, u1_ml100k, measures=["MAE"], cv=PredefinedKFold()
    )["test_mae"]

    assert mae == rs.best_score["mae"]


def test_randomizedsearchcv_same_splits():
    """Ensure that all parameter combinations are tested on the same splits (we
    check their RMSE scores are the same once averaged over the splits, which
    should be enough). We use as much parallelism as possible."""

    data_file = os.path.join(os.path.dirname(__file__), "./u1_ml100k_test")
    data = Dataset.load_from_file(data_file, reader=Reader("ml-100k"))
    kf = KFold(3, shuffle=True, random_state=4)

    # all RMSE should be the same (as param combinations are the same)
    param_distributions = {
        "n_epochs": [5],
        "lr_all": uniform(0.2, 0),
        "reg_all": uniform(0.4, 0),
        "n_factors": [5],
        "random_state": [0],
    }
    rs = RandomizedSearchCV(
        SVD, param_distributions, measures=["RMSE"], cv=kf, n_jobs=1
    )
    rs.fit(data)

    rmse_scores = [m for m in rs.cv_results["mean_test_rmse"]]
    assert len(set(rmse_scores)) == 1  # assert rmse_scores are all equal

    # Note: actually, even when setting random_state=None in kf, the same folds
    # are used because we use product(param_comb, kf.split(...)). However, it's
    # needed to have the same folds when calling fit again:
    rs.fit(data)
    rmse_scores += [m for m in rs.cv_results["mean_test_rmse"]]
    assert len(set(rmse_scores)) == 1  # assert rmse_scores are all equal


def test_randomizedsearchcv_cv_results():
    """Test the cv_results attribute"""

    f = os.path.join(os.path.dirname(__file__), "./u1_ml100k_test")
    data = Dataset.load_from_file(f, Reader("ml-100k"))
    kf = KFold(3, shuffle=True, random_state=4)
    param_distributions = {
        "n_epochs": [5],
        "lr_all": uniform(0.2, 0.3),
        "reg_all": uniform(0.4, 0.3),
        "n_factors": [5],
        "random_state": [0],
    }
    n_iter = 5
    rs = RandomizedSearchCV(
        SVD,
        param_distributions,
        n_iter=n_iter,
        measures=["RMSE", "mae"],
        cv=kf,
        return_train_measures=True,
    )
    rs.fit(data)

    # test keys split*_test_rmse, mean and std dev.
    assert rs.cv_results["split0_test_rmse"].shape == (n_iter,)
    assert rs.cv_results["split1_test_rmse"].shape == (n_iter,)
    assert rs.cv_results["split2_test_rmse"].shape == (n_iter,)
    assert rs.cv_results["mean_test_rmse"].shape == (n_iter,)
    assert np.allclose(
        rs.cv_results["mean_test_rmse"],
        np.mean(
            [
                rs.cv_results["split0_test_rmse"],
                rs.cv_results["split1_test_rmse"],
                rs.cv_results["split2_test_rmse"],
            ],
            axis=0,
        ),
    )
    assert np.allclose(
        rs.cv_results["std_test_rmse"],
        np.std(
            [
                rs.cv_results["split0_test_rmse"],
                rs.cv_results["split1_test_rmse"],
                rs.cv_results["split2_test_rmse"],
            ],
            axis=0,
        ),
    )

    # test keys split*_train_mae, mean and std dev.
    assert rs.cv_results["split0_train_rmse"].shape == (n_iter,)
    assert rs.cv_results["split1_train_rmse"].shape == (n_iter,)
    assert rs.cv_results["split2_train_rmse"].shape == (n_iter,)
    assert rs.cv_results["mean_train_rmse"].shape == (n_iter,)
    assert np.allclose(
        rs.cv_results["mean_train_rmse"],
        np.mean(
            [
                rs.cv_results["split0_train_rmse"],
                rs.cv_results["split1_train_rmse"],
                rs.cv_results["split2_train_rmse"],
            ],
            axis=0,
        ),
    )
    assert np.allclose(
        rs.cv_results["std_train_rmse"],
        np.std(
            [
                rs.cv_results["split0_train_rmse"],
                rs.cv_results["split1_train_rmse"],
                rs.cv_results["split2_train_rmse"],
            ],
            axis=0,
        ),
    )

    # test fit and train times dimensions.
    assert rs.cv_results["mean_fit_time"].shape == (n_iter,)
    assert rs.cv_results["std_fit_time"].shape == (n_iter,)
    assert rs.cv_results["mean_test_time"].shape == (n_iter,)
    assert rs.cv_results["std_test_time"].shape == (n_iter,)

    assert rs.cv_results["params"] is rs.param_combinations

    # assert that best parameter in rs.cv_results['rank_test_measure'] is
    # indeed the best_param attribute.
    best_index = np.argmin(rs.cv_results["rank_test_rmse"])
    assert rs.cv_results["params"][best_index] == rs.best_params["rmse"]
    best_index = np.argmin(rs.cv_results["rank_test_mae"])
    assert rs.cv_results["params"][best_index] == rs.best_params["mae"]


def test_randomizedsearchcv_refit(u1_ml100k):
    """Test refit method of RandomizedSearchCV class."""

    data_file = os.path.join(os.path.dirname(__file__), "./u1_ml100k_test")
    data = Dataset.load_from_file(data_file, Reader("ml-100k"))

    param_distributions = {
        "n_epochs": [5],
        "lr_all": uniform(0.002, 0.003),
        "reg_all": uniform(0.4, 0.2),
        "n_factors": [2],
    }

    # assert rs.fit() and rs.test will use best estimator for mae (first
    # appearing in measures)
    rs = RandomizedSearchCV(
        SVD, param_distributions, measures=["mae", "rmse"], cv=2, refit=True
    )
    rs.fit(data)
    rs_preds = rs.test(data.construct_testset(data.raw_ratings))
    mae_preds = rs.best_estimator["mae"].test(data.construct_testset(data.raw_ratings))
    assert rs_preds == mae_preds

    # assert rs.fit() and rs.test will use best estimator for rmse
    rs = RandomizedSearchCV(
        SVD, param_distributions, measures=["mae", "rmse"], cv=2, refit="rmse"
    )
    rs.fit(data)
    rs_preds = rs.test(data.construct_testset(data.raw_ratings))
    rmse_preds = rs.best_estimator["rmse"].test(
        data.construct_testset(data.raw_ratings)
    )
    assert rs_preds == rmse_preds
    # test that predict() can be called
    rs.predict(2, 4)

    # assert test() and predict() cannot be used when refit is false
    rs = RandomizedSearchCV(
        SVD, param_distributions, measures=["mae", "rmse"], cv=2, refit=False
    )
    rs.fit(data)
    with pytest.raises(ValueError):
        rs.test(data.construct_testset(data.raw_ratings))
    with pytest.raises(ValueError):
        rs.predict("1", "2")

    # test that error is raised if used with load_from_folds
    rs = RandomizedSearchCV(
        SVD, param_distributions, measures=["mae", "rmse"], cv=2, refit=True
    )
    with pytest.raises(ValueError):
        rs.fit(u1_ml100k)



"""
The :mod:`dataset <surprise.dataset>` module defines the :class:`Dataset` class
and other subclasses which are used for managing datasets.

Users may use both *built-in* and user-defined datasets (see the
:ref:`getting_started` page for examples). Right now, three built-in datasets
are available:

* The `movielens-100k <https://grouplens.org/datasets/movielens/>`_ dataset.
* The `movielens-1m <https://grouplens.org/datasets/movielens/>`_ dataset.
* The `Jester <https://eigentaste.berkeley.edu/dataset/>`_ dataset 2.

Built-in datasets can all be loaded (or downloaded if you haven't already)
using the :meth:`Dataset.load_builtin` method.
Summary:

.. autosummary::
    :nosignatures:

    Dataset.load_builtin
    Dataset.load_from_file
    Dataset.load_from_folds
"""

import itertools
import os
import sys
from collections import defaultdict

from .builtin_datasets import BUILTIN_DATASETS, download_builtin_dataset
from .reader import Reader
from .trainset import Trainset


class Dataset:
    """Base class for loading datasets.

    Note that you should never instantiate the :class:`Dataset` class directly
    (same goes for its derived classes), but instead use one of the three
    available methods for loading datasets."""

    def __init__(self, reader):

        self.reader = reader

    @classmethod
    def load_builtin(cls, name="ml-100k", prompt=True):
        """Load a built-in dataset.

        If the dataset has not already been loaded, it will be downloaded and
        saved. You will have to split your dataset using the :meth:`split
        <DatasetAutoFolds.split>` method. See an example in the :ref:`User
        Guide <cross_validate_example>`.

        Args:
            name(:obj:`string`): The name of the built-in dataset to load.
                Accepted values are 'ml-100k', 'ml-1m', and 'jester'.
                Default is 'ml-100k'.
            prompt(:obj:`bool`): Prompt before downloading if dataset is not
                already on disk.
                Default is True.

        Returns:
            A :obj:`Dataset` object.

        Raises:
            ValueError: If the ``name`` parameter is incorrect.
        """

        try:
            dataset = BUILTIN_DATASETS[name]
        except KeyError:
            raise ValueError(
                "unknown dataset "
                + name
                + ". Accepted values are "
                + ", ".join(BUILTIN_DATASETS.keys())
                + "."
            )

        # if dataset does not exist, offer to download it
        if not os.path.isfile(dataset.path):
            answered = not prompt
            while not answered:
                print(
                    "Dataset " + name + " could not be found. Do you want "
                    "to download it? [Y/n] ",
                    end="",
                )
                choice = input().lower()

                if choice in ["yes", "y", "", "omg this is so nice of you!!"]:
                    answered = True
                elif choice in ["no", "n", "hell no why would i want that?!"]:
                    answered = True
                    print("Ok then, I'm out!")
                    sys.exit()

            download_builtin_dataset(name)

        reader = Reader(**dataset.reader_params)

        return cls.load_from_file(file_path=dataset.path, reader=reader)

    @classmethod
    def load_from_file(cls, file_path, reader):
        """Load a dataset from a (custom) file.

        Use this if you want to use a custom dataset and all of the ratings are
        stored in one file. You will have to split your dataset using the
        :meth:`split <DatasetAutoFolds.split>` method. See an example in the
        :ref:`User Guide <load_from_file_example>`.


        Args:
            file_path(:obj:`string`): The path to the file containing ratings.
            reader(:obj:`Reader <surprise.reader.Reader>`): A reader to read
                the file.
        """

        return DatasetAutoFolds(ratings_file=file_path, reader=reader)

    @classmethod
    def load_from_folds(cls, folds_files, reader):
        """Load a dataset where folds (for cross-validation) are predefined by
        some files.

        The purpose of this method is to cover a common use case where a
        dataset is already split into predefined folds, such as the
        movielens-100k dataset which defines files u1.base, u1.test, u2.base,
        u2.test, etc... It can also be used when you don't want to perform
        cross-validation but still want to specify your training and testing
        data (which comes down to 1-fold cross-validation anyway). See an
        example in the :ref:`User Guide <load_from_folds_example>`.


        Args:
            folds_files(:obj:`iterable` of :obj:`tuples`): The list of the
                folds. A fold is a tuple of the form ``(path_to_train_file,
                path_to_test_file)``.
            reader(:obj:`Reader <surprise.reader.Reader>`): A reader to read
                the files.

        """

        return DatasetUserFolds(folds_files=folds_files, reader=reader)

    @classmethod
    def load_from_df(cls, df, reader):
        """Load a dataset from a pandas dataframe.

        Use this if you want to use a custom dataset that is stored in a pandas
        dataframe. See the :ref:`User Guide<load_from_df_example>` for an
        example.

        Args:
            df(`Dataframe`): The dataframe containing the ratings. It must have
                three columns, corresponding to the user (raw) ids, the item
                (raw) ids, and the ratings, in this order.
            reader(:obj:`Reader <surprise.reader.Reader>`): A reader to read
                the file. Only the ``rating_scale`` field needs to be
                specified.
        """

        return DatasetAutoFolds(reader=reader, df=df)

    def read_ratings(self, file_name):
        """Return a list of ratings (user, item, rating, timestamp) read from
        file_name"""

        with open(os.path.expanduser(file_name)) as f:
            raw_ratings = [
                self.reader.parse_line(line)
                for line in itertools.islice(f, self.reader.skip_lines, None)
            ]
        return raw_ratings

    def construct_trainset(self, raw_trainset):

        raw2inner_id_users = {}
        raw2inner_id_items = {}

        current_u_index = 0
        current_i_index = 0

        ur = defaultdict(list)
        ir = defaultdict(list)

        # user raw id, item raw id, translated rating, time stamp
        for urid, irid, r, timestamp in raw_trainset:
            try:
                uid = raw2inner_id_users[urid]
            except KeyError:
                uid = current_u_index
                raw2inner_id_users[urid] = current_u_index
                current_u_index += 1
            try:
                iid = raw2inner_id_items[irid]
            except KeyError:
                iid = current_i_index
                raw2inner_id_items[irid] = current_i_index
                current_i_index += 1

            ur[uid].append((iid, r))
            ir[iid].append((uid, r))

        n_users = len(ur)  # number of users
        n_items = len(ir)  # number of items
        n_ratings = len(raw_trainset)

        trainset = Trainset(
            ur,
            ir,
            n_users,
            n_items,
            n_ratings,
            self.reader.rating_scale,
            raw2inner_id_users,
            raw2inner_id_items,
        )

        return trainset

    def construct_testset(self, raw_testset):

        return [(ruid, riid, r_ui_trans) for (ruid, riid, r_ui_trans, _) in raw_testset]


class DatasetUserFolds(Dataset):
    """A derived class from :class:`Dataset` for which folds (for
    cross-validation) are predefined."""

    def __init__(self, folds_files=None, reader=None):

        Dataset.__init__(self, reader)
        self.folds_files = folds_files

        # check that all files actually exist.
        for train_test_files in self.folds_files:
            for f in train_test_files:
                if not os.path.isfile(os.path.expanduser(f)):
                    raise ValueError("File " + str(f) + " does not exist.")


class DatasetAutoFolds(Dataset):
    """A derived class from :class:`Dataset` for which folds (for
    cross-validation) are not predefined. (Or for when there are no folds at
    all)."""

    def __init__(self, ratings_file=None, reader=None, df=None):

        Dataset.__init__(self, reader)
        self.has_been_split = False  # flag indicating if split() was called.

        if ratings_file is not None:
            self.ratings_file = ratings_file
            self.raw_ratings = self.read_ratings(self.ratings_file)
        elif df is not None:
            self.df = df
            self.raw_ratings = [
                (uid, iid, float(r), None)
                for (uid, iid, r) in self.df.itertuples(index=False)
            ]
        else:
            raise ValueError("Must specify ratings file or dataframe.")

    def build_full_trainset(self):
        """Do not split the dataset into folds and just return a trainset as
        is, built from the whole dataset.

        User can then query for predictions, as shown in the :ref:`User Guide
        <train_on_whole_trainset>`.

        Returns:
            The :class:`Trainset <surprise.Trainset>`.
        """

        return self.construct_trainset(self.raw_ratings)



"""
This module descibes how to test the performances of an algorithm on the
trainset.
"""

from surprise import accuracy, Dataset, SVD
from surprise.model_selection import KFold

data = Dataset.load_builtin("ml-100k")

algo = SVD()

trainset = data.build_full_trainset()
algo.fit(trainset)

testset = trainset.build_testset()
predictions = algo.test(testset)
# RMSE should be low as we are biased
accuracy.rmse(predictions, verbose=True)  # ~ 0.68 (which is low)

# We can also do this during a cross-validation procedure!
print("CV procedure:")

kf = KFold(n_splits=3)
for i, (trainset_cv, testset_cv) in enumerate(kf.split(data)):
    print("fold number", i + 1)
    algo.fit(trainset_cv)

    print("On testset,", end="  ")
    predictions = algo.test(testset_cv)
    accuracy.rmse(predictions, verbose=True)

    print("On trainset,", end=" ")
    predictions = algo.test(trainset_cv.build_testset())
    accuracy.rmse(predictions, verbose=True)



"""
This module descibes how to build your own prediction algorithm. Please refer
to User Guide for more insight.
"""

from surprise import AlgoBase, Dataset
from surprise.model_selection import cross_validate


class MyOwnAlgorithm(AlgoBase):
    def __init__(self):

        # Always call base method before doing anything.
        AlgoBase.__init__(self)

    def estimate(self, u, i):

        return 3


data = Dataset.load_builtin("ml-100k")
algo = MyOwnAlgorithm()

cross_validate(algo, data, verbose=True)



"""
This module illustrates how to retrieve the top-10 items with highest rating
prediction. We first train an SVD algorithm on the MovieLens dataset, and then
predict all the ratings for the pairs (user, item) that are not in the training
set. We then retrieve the top-10 prediction for each user.
"""

from collections import defaultdict

from surprise import Dataset, SVD


def get_top_n(predictions, n=10):
    """Return the top-N recommendation for each user from a set of predictions.

    Args:
        predictions(list of Prediction objects): The list of predictions, as
            returned by the test method of an algorithm.
        n(int): The number of recommendation to output for each user. Default
            is 10.

    Returns:
    A dict where keys are user (raw) ids and values are lists of tuples:
        [(raw item id, rating estimation), ...] of size n.
    """

    # First map the predictions to each user.
    top_n = defaultdict(list)
    for uid, iid, true_r, est, _ in predictions:
        top_n[uid].append((iid, est))

    # Then sort the predictions for each user and retrieve the k highest ones.
    for uid, user_ratings in top_n.items():
        user_ratings.sort(key=lambda x: x[1], reverse=True)
        top_n[uid] = user_ratings[:n]

    return top_n


# First train an SVD algorithm on the movielens dataset.
data = Dataset.load_builtin("ml-100k")
trainset = data.build_full_trainset()
algo = SVD()
algo.fit(trainset)

# Than predict ratings for all pairs (u, i) that are NOT in the training set.
testset = trainset.build_anti_testset()
predictions = algo.test(testset)

top_n = get_top_n(predictions, n=10)

# Print the recommended items for each user
for uid, user_ratings in top_n.items():
    print(uid, [iid for (iid, _) in user_ratings])



from . import dump, model_selection
from .builtin_datasets import get_dataset_dir
from .dataset import Dataset
from .prediction_algorithms import (
    AlgoBase,
    BaselineOnly,
    CoClustering,
    KNNBaseline,
    KNNBasic,
    KNNWithMeans,
    KNNWithZScore,
    NMF,
    NormalPredictor,
    Prediction,
    PredictionImpossible,
    SlopeOne,
    SVD,
    SVDpp,
)
from .reader import Reader
from .trainset import Trainset

__all__ = [
    "AlgoBase",
    "NormalPredictor",
    "BaselineOnly",
    "KNNBasic",
    "KNNWithMeans",
    "KNNBaseline",
    "SVD",
    "SVDpp",
    "NMF",
    "SlopeOne",
    "CoClustering",
    "PredictionImpossible",
    "Prediction",
    "Dataset",
    "Reader",
    "Trainset",
    "dump",
    "KNNWithZScore",
    "get_dataset_dir",
    "model_selection",
]

__version__ = "1.2.6"
